@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2016 Codership Oy <info@codership.com>
+ * Copyright (C) 2010-2018 Codership Oy <info@codership.com>
  */
 
 #include "gcache_rb_store.hpp"
@@ -10,6 +10,8 @@
 #include <gu_logger.hpp>
 #include <gu_throw.hpp>
 #include <gu_progress.hpp>
+#include <gu_hexdump.hpp>
+#include <gu_hash.h>
 
 #include <cassert>
 
@@ -45,6 +47,7 @@ namespace gcache
                             size_t             size,
                             seqno2ptr_t&       seqno2ptr,
                             gu::UUID&          gid,
+                            int const          dbg,
                             bool const         recover)
     :
         fd_        (name, check_size(size)),
@@ -63,8 +66,10 @@ namespace gcache
         size_trail_(0),
 //        mallocs_   (0),
 //        reallocs_  (0),
+        debug_     (dbg & DEBUG),
         open_      (true)
     {
+        assert((uintptr_t(start_) % MemOps::ALIGNMENT) == 0);
         constructor_common ();
         open_preamble(recover);
         BH_clear (BH_cast(next_));
@@ -138,6 +143,7 @@ namespace gcache
     BufferHeader*
     RingBuffer::get_new_buffer (size_type const size)
     {
+        assert((size % MemOps::ALIGNMENT) == 0);
         assert_size_free();
 
         BH_assert_clear(BH_cast(next_));
@@ -230,6 +236,7 @@ namespace gcache
 #endif
 
     found_space:
+        assert((uintptr_t(ret) % MemOps::ALIGNMENT) == 0);
         size_used_ += size;
         assert (size_used_ <= size_cache_);
         assert (size_free_ >= size);
@@ -244,6 +251,7 @@ namespace gcache
         bh->ctx     = this;
 
         next_ = ret + size;
+        assert((uintptr_t(next_) % MemOps::ALIGNMENT) == 0);
         assert (next_ + sizeof(BufferHeader) <= end_);
         BH_clear (BH_cast(next_));
         assert_sizes();
@@ -533,7 +541,6 @@ namespace gcache
     void
     RingBuffer::write_preamble(bool const synced)
     {
-        static int const VERSION(1);
         uint8_t* const preamble(reinterpret_cast<uint8_t*>(preamble_));
 
         std::ostringstream os;
@@ -576,8 +583,8 @@ namespace gcache
     void
     RingBuffer::open_preamble(bool const do_recover)
     {
+        int version(0); // used only for recovery on upgrade
         uint8_t* const preamble(reinterpret_cast<uint8_t*>(preamble_));
-        int version(0);
         long long seqno_max(SEQNO_ILL);
         long long seqno_min(SEQNO_ILL);
         off_t offset(-1);
@@ -614,7 +621,9 @@ namespace gcache
            version = 0;
         }
 
-        if (offset < -1 || (preamble + offset + sizeof(BufferHeader)) > end_)
+        if (offset < -1 ||
+            (preamble + offset + sizeof(BufferHeader)) > end_ ||
+            (version >= 2 && offset >= 0 && (offset % MemOps::ALIGNMENT)))
         {
            log_warn << "Bogus offset in GCache ring buffer preamble: "
                     << offset << ". Assuming unknown.";
@@ -630,7 +639,7 @@ namespace gcache
 
                 try
                 {
-                    recover(offset - (start_ - preamble));
+                    recover(offset - (start_ - preamble), version);
                 }
                 catch (gu::Exception& e)
                 {
@@ -641,7 +650,7 @@ namespace gcache
             }
             else
             {
-                log_warn << "Skipped GCache ring buffer recovery: could not "
+                log_info << "Skipped GCache ring buffer recovery: could not "
                     "determine history UUID.";
             }
         }
@@ -656,7 +665,7 @@ namespace gcache
     }
 
     int64_t
-    RingBuffer::scan(off_t const offset)
+    RingBuffer::scan(off_t const offset, int const scan_step)
     {
         int segment_scans(0);
         int64_t seqno_max(-1);
@@ -665,11 +674,14 @@ namespace gcache
         size_t collision_count(0);
         int64_t erase_up_to(-1);
         uint8_t* segment_start(start_);
+        uint8_t* segment_end(end_ - sizeof(BufferHeader));
 
         /* start at offset (first segment) if we know it and it is valid */
         if (offset >= 0)
         {
-            if (start_ + offset + 2*sizeof(BufferHeader) < end_)
+            assert(0 == (offset % scan_step));
+
+            if (start_ + offset + sizeof(BufferHeader) < segment_end)
                 /* we know exaclty where the first segment starts */
                 segment_start = start_ + offset;
             else
@@ -677,11 +689,8 @@ namespace gcache
                 segment_scans = 1;
         }
 
-        gu::Progress<ptrdiff_t> progress("GCache::RingBuffer initial scan ",
-                                         " bytes",
-                                         "PT5S", /* 5 sec */
-                                         1<<22,  /* 4Mb */
-                                         end_ - start_);
+        gu::Progress<ptrdiff_t> progress("GCache::RingBuffer initial scan",
+                                         " bytes", end_ - start_, 1<<22 /*4Mb*/);
 
         while (segment_scans < 2)
         {
@@ -692,11 +701,13 @@ namespace gcache
 
 #define GCACHE_SCAN_BUFFER_TEST                                 \
             (BH_test(bh) && bh->size > 0 &&                     \
-             ptr + bh->size + sizeof(BufferHeader) <= end_ &&   \
+             ptr + bh->size <= segment_end &&                   \
              BH_test(BH_cast(ptr + bh->size)))
 
             while (GCACHE_SCAN_BUFFER_TEST)
             {
+                assert((uintptr_t(bh) % scan_step) == 0);
+
                 bh->flags |= BUFFER_RELEASED;
                 bh->ctx    = this;
 
@@ -717,22 +728,69 @@ namespace gcache
                     {
                         collision_count++;
 
-                        log_info <<"Attempt to reuse the same seqno: " << seqno_g
-                                 << ". New ptr = " << static_cast<void*>(bh+1)
-                                 << ", previous ptr = " << res.first->second;
+                        /* compare two buffers */
+                        const void* const old_ptr(res.first->second);
+                        BufferHeader* const old_bh
+                            (old_ptr ? ptr2BH(old_ptr) : NULL);
+
+                        bool const same_meta(NULL != old_bh &&
+                            bh->seqno_g == old_bh->seqno_g  &&
+                            bh->size    == old_bh->size     &&
+                            bh->flags   == old_bh->flags);
+
+                        const void* const new_ptr(static_cast<void*>(bh+1));
+
+                        uint8_t cs_old[16] = { 0, };
+                        uint8_t cs_new[16] = { 0, };
+                        if (same_meta)
+                        {
+                            gu_fast_hash128(old_ptr,
+                                            old_bh->size - sizeof(BufferHeader),
+                                            cs_old);
+                            gu_fast_hash128(new_ptr,
+                                            bh->size - sizeof(BufferHeader),
+                                            cs_new);
+                        }
+
+                        bool const same_data(same_meta &&
+                                             !::memcmp(cs_old, cs_new,
+                                                       sizeof(cs_old)));
+                        std::ostringstream msg;
+
+                        msg << "Attempt to reuse the same seqno: " << seqno_g
+                            << ". New ptr = " << new_ptr << ", " << bh
+                            << ", cs: " << gu::Hexdump(cs_new, sizeof(cs_new))
+                            << ", previous ptr = " << res.first->second;
+
                         empty_buffer(bh); // this buffer is unusable
                         assert(BH_is_released(bh));
 
-                        if (res.first->second != NULL)
+                        if (old_bh != NULL)
                         {
-                            BufferHeader* b(ptr2BH(res.first->second));
-                            empty_buffer(b);
-                            assert(BH_is_released(b));
-                            res.first->second = NULL; // mark entry as invalid
-                            // but don't remove from the map to block this seqno
+                            msg << ", " << old_bh << ", cs: "
+                                << gu::Hexdump(cs_old,sizeof(cs_old));
+
+                            if (!same_data) // no way to choose which is correct
+                            {
+                                empty_buffer(old_bh);
+                                assert(BH_is_released(old_bh));
+                                res.first->second = NULL;
+                                /* mark entry as invalid, but don't remove from
+                                 * the map to block this seqno */
+
+                                if (erase_up_to < seqno_g) erase_up_to = seqno_g;
+                            }
                         }
 
-                        if (erase_up_to < seqno_g) erase_up_to = seqno_g;
+                        log_info << msg.str();
+
+                        if (same_data) {
+                            log_info << "Contents are the same, discarding "
+                                     << new_ptr;
+                        } else {
+                            assert(NULL == res.first->second);
+                            log_info << "Contents differ. Discarding both.";
+                        }
                     }
                 }
 
@@ -743,15 +801,15 @@ namespace gcache
 
             if (!BH_is_clear(bh))
             {
-                if (start_ == segment_start)
+                if (start_ == segment_start && ptr != first_)
                 {
                     log_warn << "Failed to scan the last segment to the end. "
                             "Last events may be missing. Last recovered event: "
                              << gid_ << ':' << seqno_max;
                 }
 
-                /* do best effort */
-                BH_clear(bh);
+                /* end of file, do best effort */
+                if (end_ - sizeof(BufferHeader) == segment_end) BH_clear(bh);
             }
 
             if (offset > 0 && segment_start == start_ + offset)
@@ -760,6 +818,7 @@ namespace gcache
                 assert(1 == segment_scans);
                 first_ = segment_start;
                 size_trail_ = end_ - ptr;
+                segment_end = segment_start;
                 segment_start = start_;
             }
             else if (offset < 0 && segment_start == start_)
@@ -771,10 +830,9 @@ namespace gcache
                 while (!GCACHE_SCAN_BUFFER_TEST &&
                        ptr + sizeof(BufferHeader) < end_)
                 {
-                    progress.update(1);
-                    ptr += 1;
+                    progress.update(scan_step);
+                    ptr += scan_step;
                     bh = BH_cast(ptr);
-                    /* it would have been nice to use alignment of 8 or so */
                 }
 
                 if (GCACHE_SCAN_BUFFER_TEST)
@@ -782,7 +840,7 @@ namespace gcache
                     segment_start = ptr;
                     first_ = segment_start;
                 }
-                else if (ptr + sizeof(BufferHeader) == end_)
+                else if (ptr + sizeof(BufferHeader) >= end_)
                 {
                     /* perhaps it was a single segment starting at start_ */
                     first_ = start_;
@@ -816,7 +874,7 @@ namespace gcache
                     /* first (end) segment was scanned last, estimate trail */
                     size_trail_ = end_ - ptr;
                 }
-                else if (offset > 0 && next_ >= first_)
+                else if (offset > 0 && next_ > first_)
                 {
                     size_trail_ = 0;
                 }
@@ -830,18 +888,17 @@ namespace gcache
     }
 
     void
-    RingBuffer::recover(off_t const offset)
+    RingBuffer::recover(off_t const offset, int version)
     {
         static const char* const diag_prefix = "Recovering GCache ring buffer: ";
 
         /* scan the buffer and populate seqno2ptr map */
-        int64_t const lower(scan(offset));
+        int64_t const lower(scan(offset, version > 0 ? MemOps::ALIGNMENT : 1));
 
         if (!seqno2ptr_.empty())
         {
-            assert(first_ != next_);
-            assert(next_ < first_ || size_trail_ == 0);
-            assert(next_ > first_ || size_trail_ >  0);
+            assert(next_ <= first_ || size_trail_ == 0);
+            assert(next_ >  first_ || size_trail_ >  0);
 
             /* find the last gapless seqno sequence */
             seqno2ptr_t::reverse_iterator r(seqno2ptr_.rbegin());
@@ -890,7 +947,7 @@ namespace gcache
                 /* clear up seqno2ptr map */
                 for (; r != seqno2ptr_.rend(); ++r)
                 {
-                    empty_buffer(ptr2BH(r->second));
+                    if (r->second) empty_buffer(ptr2BH(r->second));
                 }
                 seqno2ptr_.erase(seqno2ptr_.begin(), seqno2ptr_.find(seqno_min));
             }
@@ -912,8 +969,8 @@ namespace gcache
 
             /* trim next_: start with the last seqno and scan forward up to the
              * current next_. Update to the end of the last non-empty buffer. */
-            BufferHeader* last_bh(NULL);
             bh = ptr2BH(seqno2ptr_[seqno_max]);
+            BufferHeader* last_bh(bh);
             while (bh != BH_cast(next_))
             {
                 if (gu_likely(bh->size) > 0)
@@ -930,9 +987,24 @@ namespace gcache
                     bh = BH_cast(start_); // rollover
                 }
             }
-            assert(last_bh);
             next_ = reinterpret_cast<uint8_t*>(BH_next(last_bh));
 
+            /* Even if previous buffers were not aligned, make sure from
+             * now on they are - adjust next_ pointer and last buffer size */
+            if (uintptr_t(next_) % MemOps::ALIGNMENT)
+            {
+                uint8_t* const n(MemOps::align_ptr(next_));
+                assert(n > next_);
+                size_type const size_diff(n - next_);
+                assert(size_diff < MemOps::ALIGNMENT);
+                assert(last_bh->size > 0);
+                last_bh->size += size_diff;
+                next_ = n;
+                assert(BH_next(last_bh) == BH_cast(next_));
+            }
+            assert((uintptr_t(next_) % MemOps::ALIGNMENT) == 0);
+            BH_clear(BH_cast(next_));
+
             /* at this point we must have at least one seqno'd buffer */
             assert(next_ != first_);
 
@@ -944,11 +1016,8 @@ namespace gcache
 
             /* now discard all the locked-in buffers (see seqno_reset()) */
             gu::Progress<size_t> progress(
-                "GCache::RingBuffer unused buffers scan ",
-                " bytes",
-                "PT5S", /* 5 sec */
-                1<<22,  /* 4Mb   */
-                size_used_);
+                "GCache::RingBuffer unused buffers scan",
+                " bytes", size_used_, 1<<22 /* 4Mb */);
 
             size_t total(0);
             size_t locked(0);
